
/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or 
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 */

/*
 *  $Id: usb-xboxir.c,v 1.3 2002/12/02 23:04:58 oliverschwartz Exp $
 *
 *  Copyright (c) 2002 Steven Toth <steve@toth.demon.co.uk>
 *
 *  XBOX DVD dongle infrared device driver for the input driver suite.
 *
 *  This work was derived from the usbkbd.c kernel module.
 *
 *  History:
 *
 *  2002_08_31 - 0.1 - Initial release
 *  2002_09_02 - 0.2 - Added IOCTL support enabling user space administration
 *                     of the translation matrix.
 *
 */

#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/module.h>
#include <linux/proc_fs.h>
#include <linux/input.h>
#include <linux/init.h>
#include <linux/usb.h>
#include <asm/uaccess.h>
#include "usb-xboxir.h"


#ifndef MODULE
#define MODULE
#endif

#define DRIVER_VERSION		"0.2"
#define DRIVER_AUTHOR			"Steven Toth <steve@toth.demon.co.uk>"
#define DRIVER_DESC				"USB HID XBOX IR driver"
#define PROC_FILE_NAME		"xboxir"
#define PROC_FILE_PERMS		S_IFREG | S_IRUGO
#define XBOXIR_MINOR			240
#define IR_INTERVAL		300

#ifdef dbg
#undef dbg
#define dbg(format, arg...) do { if(xboxir_debug) printk(KERN_INFO __FILE__ ": " format "\n" , ## arg); } while (0)

unsigned int xboxir_debug = 1;
struct proc_dir_entry *proc_file_ent;


/*  Dump the contents of the translation table in ASCII to the /proc file */
static int usb_xboxir_read_proc_ir( char *buf, char **start, off_t offset, int count, int *eof, void *data)
{
	int i=0, len=0;

	if(count>=512) {
	
		len += sprintf(buf+len,"XBOX:HID\n");
		while(rc_kbd_matrix[i].rc_code != 0) {
			len += sprintf(buf+len,"0x%02x:0x%04x\n"
				,rc_kbd_matrix[i].rc_code
				,rc_kbd_matrix[i].kbd_code);
			i++;
		}

	} else {
		dbg("user attemped to read /proc but user bufferlen was < 512 chars, not big enough, increase");
	}

	*eof = 1;
	return len;
}

/* Perform an lookup for xbox rc_code in the table, return the kbd_code. */
static unsigned int get_kdb_code(unsigned char c)
{
	int i=0;

	/* the table is order so we can take a short cut for invalid code numbers */
	while(rc_kbd_matrix[i].rc_code != 0) {
		if(rc_kbd_matrix[i].rc_code > c) return 0; /* wasn't in list */
		if(rc_kbd_matrix[i].rc_code == c)
			return rc_kbd_matrix[i].kbd_code;
		i++;
	}

	return 0;
}

/* Perform an lookup for xbox rc_code in the table, set the kbd_code. */
static unsigned int set_kdb_code(unsigned char c, unsigned int new_code)
{
	int i=0;

	/* the table is order so we can take a short cut for invalid code numbers */
	while(rc_kbd_matrix[i].rc_code != 0) {
		if(rc_kbd_matrix[i].rc_code > c) return 0; /* wasn't in list */
		if(rc_kbd_matrix[i].rc_code == c) {
			dbg("Changing translation from HID event 0x%X to 0x%X",rc_kbd_matrix[i].kbd_code, new_code);
			rc_kbd_matrix[i].kbd_code = new_code;
			return 1;
		}
		i++;
	}

	return 0;
}


/* Start of IOCTL specifics */
static int usb_xboxir_ioc_ioctl (struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
	struct t_rc_kbd_matrix input;
	int err=0;
	int ret=0;

	if (_IOC_TYPE(cmd) != XBOXIR_IOC_MAGIC) return -ENOTTY;
	if (_IOC_NR(cmd) > XBOXIR_IOC_MAXNR) return -ENOTTY;

	/* Check that the user space addresses are valid - we don't want an exception */
	if (_IOC_DIR(cmd) & _IOC_READ)
		err = !access_ok(VERIFY_WRITE, (void *)arg, _IOC_SIZE(cmd));
	else if (_IOC_DIR(cmd) & _IOC_WRITE)
		err = !access_ok(VERIFY_READ, (void *)arg, _IOC_SIZE(cmd));
	if (err) return -EFAULT;

	switch(cmd) {

		case XBOXIR_IOCSQSET:	/* user performing an update */
			if (copy_from_user(&input, (int *)arg, sizeof(input))) return -EFAULT;
			dbg("user sent struct containing rc_code=%X, kbd_code=%X",input.rc_code,input.kbd_code);
			if( (input.kbd_code = set_kdb_code( input.rc_code, input.kbd_code )) == 0 ) {
				ret = -1;
			} else {
				ret = 0;
			}

			break;

		case XBOXIR_IOCSQGET:	/* User is performing a query */

			if (copy_from_user(&input, (int *)arg, sizeof(input))) return -EFAULT;
			dbg("user sent struct containing rc_code=%X, kbd_code=%X",input.rc_code,input.kbd_code);
			if( (input.kbd_code = get_kdb_code( input.rc_code )) == 0 ) {
				ret = -1;
			} else {
				/* Return the structure to the user */
				dbg("we return containing rc_code=%X, kbd_code=%X",input.rc_code,input.kbd_code);
				if (copy_to_user((int *)arg, &input, sizeof(input))) return -EFAULT;
				ret = 0;
			}
			break;
		default:
			return -ENOTTY;
	}

	return ret;
}

static int usb_xboxir_ioc_open (struct inode *inode, struct file *file)
{
	return 0;
}

/* End of IOCTL specifics */

/*  USB callback completion handler
 *  Code in transfer_buffer is received as six unsigned chars
 *  Example PLAY=00 06 ea 0a 40 00
 *  The command is located in byte[2], the rest are ignored.
 *  Key position is byte[4] bit0 (7-0 format) 0=down, 1=up
 *  All other bits are unknown / now required.
 */
static void usb_xboxir_irq(struct urb *urb)
{
	struct timeval key_time;
	struct usb_xboxir *xir = urb->context;
	unsigned int kbd_code=0;
	unsigned int key_direction=1; // 0=up, 1=down
	signed int time_diff=0;

	// get current time
	do_gettimeofday(&key_time);

	if (urb->status) return;
	if (urb->actual_length < 6) return;

	/* Messy/unnecessary, fix this */
	memcpy(xir->irpkt, urb->transfer_buffer, 6);
	if ( (kbd_code = get_kdb_code( xir->irpkt[2] )) == 0) return;

	/* Set the key action based in the sent action */
	key_direction =  ( xir->irpkt[4] & 1 ? 1 : 0) ;

	dbg("-------- KEY READ : %d KEY_DIRECTION : %d",kbd_code,key_direction);
	
	// if key press is different to previous then always report it
	if (xir->previous_kbd_code != kbd_code)
	{
		// send key down and up events
		input_report_key(&xir->dev, kbd_code, 0);
		input_report_key(&xir->dev, kbd_code, 1);

		dbg("-------- KEY SENT : %d KEY_DIRECTION : %d",kbd_code, 0);
		dbg("-------- KEY SENT : %d KEY_DIRECTION : %d",kbd_code, 1);
		
		// remember details of key press
		xir->previous_kbd_code = kbd_code;
		xir->previous_kbd_direction = key_direction;
		xir->previous_kbd_time = key_time;
	}
	// otherwise only report key if IR_INTERVAL has elapsed
	else
	{
		// calculate time difference between previous and current in milliseconds
		time_diff = (key_time.tv_sec - xir->previous_kbd_time.tv_sec) * 1000;
		time_diff += (key_time.tv_usec - xir->previous_kbd_time.tv_usec) / 1000;
		
		// if time difference greater than IR interval then report key,
		// otherwise ignore it
		if (time_diff > IR_INTERVAL)
		{
			// report key presses
			input_report_key(&xir->dev, kbd_code, 0);
			input_report_key(&xir->dev, kbd_code, 1);

			dbg("-------- KEY SENT : %d KEY_DIRECTION : %d",kbd_code, 0);
			dbg("-------- KEY SENT : %d KEY_DIRECTION : %d",kbd_code, 1);

			// remember details of key press
			xir->previous_kbd_code = kbd_code;
			xir->previous_kbd_direction = key_direction;
			xir->previous_kbd_time = key_time;
		}
	}

//	dbg("usb_xboxir_irq: actual_length=%d",urb->actual_length);
//	dbg("%02x %02x %02x %02x %02x %02x"
//,xir->irpkt[0],xir->irpkt[1],xir->irpkt[2],xir->irpkt[3],xir->irpkt[4],xir->irpkt[5]);

}

static int usb_xboxir_open(struct input_dev *dev)
{
	struct usb_xboxir *xir = dev->private;

	if (xir->open++)
		return 0;

	xir->irq.dev = xir->usbdev;
	if (usb_submit_urb(&xir->irq))
		return -EIO;

	return 0;
}

static void usb_xboxir_close(struct input_dev *dev)
{
	struct usb_xboxir *xir = dev->private;

	if (!--xir->open)
		usb_unlink_urb(&xir->irq);
}

static void *usb_xboxir_probe(struct usb_device *dev, unsigned int ifnum,
			   const struct usb_device_id *id)
{
	struct usb_interface *iface;
	struct usb_interface_descriptor *interface;
	struct usb_endpoint_descriptor *endpoint;
	struct usb_xboxir *xir;
	int i, pipe, maxp;
	char *buf;

	
	dbg("usb_xboxir_probe");

	iface = &dev->actconfig->interface[ifnum];
	interface = &iface->altsetting[iface->act_altsetting];

	if (interface->bNumEndpoints != 1) return NULL;

	endpoint = interface->endpoint + 0;
	if (!(endpoint->bEndpointAddress & 0x80)) return NULL;
	if ((endpoint->bmAttributes & 3) != 3) return NULL;

	pipe = usb_rcvintpipe(dev, endpoint->bEndpointAddress);
	maxp = usb_maxpacket(dev, pipe, usb_pipeout(pipe));

	usb_set_protocol(dev, interface->bInterfaceNumber, 0);
	usb_set_idle(dev, interface->bInterfaceNumber, 0, 0);

	if (!(xir = kmalloc(sizeof(struct usb_xboxir), GFP_KERNEL))) return NULL;
	memset(xir, 0, sizeof(struct usb_xboxir));

	xir->usbdev = dev;

	// The kinds of events we can send (keyboard)
	xir->dev.evbit[0] = BIT(EV_KEY);

	i=0;
	while(rc_kbd_matrix[i].rc_code != 0) {
		set_bit(rc_kbd_matrix[i++].kbd_code, xir->dev.keybit);
	}
	clear_bit(0, xir->dev.keybit);
	
	xir->dev.private = xir;
	xir->dev.event = NULL;
	xir->dev.open = usb_xboxir_open;
	xir->dev.close = usb_xboxir_close;

	FILL_INT_URB(&xir->irq, dev, pipe, xir->irpkt, maxp > 8 ? 8 : maxp,
		usb_xboxir_irq, xir, endpoint->bInterval);

	xir->dr.bRequestType = USB_TYPE_CLASS | USB_RECIP_INTERFACE;
	xir->dr.bRequest = USB_REQ_SET_REPORT;
	xir->dr.wValue = 0x200;
	xir->dr.wIndex = interface->bInterfaceNumber;
	xir->dr.wLength = 1;

	xir->dev.name = xir->name;
	xir->dev.idbus = BUS_USB;
	xir->dev.idvendor = dev->descriptor.idVendor;
	xir->dev.idproduct = dev->descriptor.idProduct;
	xir->dev.idversion = dev->descriptor.bcdDevice;

	if (!(buf = kmalloc(63, GFP_KERNEL))) {
		kfree(xir);
		return NULL;
	}

	if (dev->descriptor.iManufacturer &&
		usb_string(dev, dev->descriptor.iManufacturer, buf, 63) > 0)
			strcat(xir->name, buf);
	
	if (dev->descriptor.iProduct &&
		usb_string(dev, dev->descriptor.iProduct, buf, 63) > 0)
			sprintf(xir->name, "%s %s", xir->name, buf);

	if (!strlen(xir->name))
		sprintf(xir->name, "USB HID XBOX IR %04x:%04x",
			xir->dev.idvendor, xir->dev.idproduct);

	kfree(buf);

	input_register_device(&xir->dev);

	dbg("input%d: %s on usb%d:%d.%d", xir->dev.number, xir->name, dev->bus->busnum, dev->devnum, ifnum);

	return xir;
}

static void usb_xboxir_disconnect(struct usb_device *dev, void *ptr)
{
	struct usb_xboxir *xir = ptr;
	usb_unlink_urb(&xir->irq);
	input_unregister_device(&xir->dev);
	kfree(xir);
}

static struct usb_device_id usb_xboxir_id_table [] = {
	{ USB_DEVICE(0x045e, 0x0284) }, /* Microsoft, DVD dongle */
	{ USB_DEVICE(0xffff, 0xffff) }, /* Generic, DVD dongl
	{ } /* Terminating entry */
};

MODULE_DEVICE_TABLE (usb, usb_xboxir_id_table);

static struct file_operations usb_xboxir_fops =
{
  owner:    THIS_MODULE,
  ioctl:    usb_xboxir_ioc_ioctl,
  open:     usb_xboxir_ioc_open,
};


static struct usb_driver usb_xboxir_driver = {
	name:				"usb-xboxir",
	probe:			usb_xboxir_probe,
	disconnect:	usb_xboxir_disconnect,
	id_table:		usb_xboxir_id_table,
	fops:				&usb_xboxir_fops,
	minor:			XBOXIR_MINOR,
};

static int __init usb_xboxir_init(void)
{
	usb_register(&usb_xboxir_driver);
	proc_file_ent = create_proc_read_entry(PROC_FILE_NAME,PROC_FILE_PERMS,NULL,usb_xboxir_read_proc_ir,NULL);
	dbg("%s:%s",DRIVER_VERSION,DRIVER_DESC);
	return 0;
}

static void __exit usb_xboxir_exit(void)
{
	if(proc_file_ent) remove_proc_entry(PROC_FILE_NAME,NULL);
	usb_deregister(&usb_xboxir_driver);
}

MODULE_PARM(xboxir_debug,"i");
MODULE_AUTHOR( DRIVER_AUTHOR );
MODULE_DESCRIPTION( DRIVER_DESC );
MODULE_LICENSE("GPL");

module_init(usb_xboxir_init);
module_exit(usb_xboxir_exit);

#undef dbg
#endif

